# ##################################################################################################
#
#  Title
#
#   langworks.auth.py
#
#  License
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
#  Description
#
#    Part of the Langworks framework, implementing various utilities for dealing with 
#    authentication against LLM providers.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
from dataclasses import (
    dataclass,
    field
)

from typing import (
    Callable
)


# Local ############################################################################################

# vectorworks.retrievers
from .middleware.generic import (
    Middleware
)


# ##################################################################################################
# Classes
# ##################################################################################################

# AuthenticationDetails ############################################################################

@dataclass
class AuthenticationDetails:

    """
    Holds authentication details that may be retrieved by an :py:type:`Authenticator`.
    """

    username : str | None = field(default = None)
    """
    Username with which to authenticate against the LLM provider.
    """

    password : str | None = field(default = None)
    """
    Password with which to authenticate against the LLM provider.
    """

    token    : str | None = field(default = None)
    """
    Access token with which to authenticate against the LLM provider.
    """

    # End of dataclass 'AuthenticationDetails' #####################################################


# ##################################################################################################
# Types
# ##################################################################################################

# Authenticator ####################################################################################

type Authenticator = Callable[[Middleware], AuthenticationDetails]
"""
Callable invoked by an :py:class:`Middleware` when seeking access to an external resource.
"""


# ##################################################################################################
# Functions
# ##################################################################################################

# UsernameCredentials ##############################################################################

def UsernameCredentials(username : str, password : str | None = None) -> AuthenticationDetails:

    """
    Generates an :py:type:`Authenticator` that generates authentication details with the given
    username and password.

    Parameters
    ----------

    username
        Username with which to authenticate against the LLM provider.

    password
        Password with which to authenticate against the LLM provider.
    """

    return lambda _: AuthenticationDetails(username, password)

    # End of function 'UsernameCredentials' ########################################################


# TokenCredentials #################################################################################

def TokenCredentials(token : str = None)  -> AuthenticationDetails:

    """
    Generates an :py:type:`Authenticator` that generates authentication details with the given
    token

    Parameters
    ----------

    token
        Token with which to authenticate against the LLM provider.
    """

    return lambda _: AuthenticationDetails(token = token)

    # End of function 'TokenCredentials' ###########################################################

# End of File ######################################################################################


 # ##################################################################################################
#
# Title:
#
#   langworks.middleware.generic.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, implementing various generic classes for use by middlewares.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
import dataclasses
from dataclasses import (
    dataclass,
    field
)

import functools

import typing
from typing import (
    Any
)


# Third-party ######################################################################################

# OpenAI API (OpenAI) [lazy]
if typing.TYPE_CHECKING:
    import openai


# Local ############################################################################################

from ..messages import (
    Thread
)


# ##################################################################################################
# Functions
# ##################################################################################################

# get_dataclass_defaults ###########################################################################

@functools.cache
def get_dataclass_defaults(cls : dataclass) -> dict[str, Any]:

    """
    Utility function used to draw up a lookup map for the default values of the fields of a 
    dataclass.
    """

    return {
        # Index by field name.
        field.name: (
            # Get default value or use default factory to create a new default instance.
            field.default 
            or (
                field.default_factory() 
                if not isinstance(field.default_factory, dataclasses._MISSING_TYPE)
                else None
            )
        )
        # Iterate over all fields.
        for field in dataclasses.fields(cls)
        # But only the fields that are initialized and comparable.
        if field.init == True and field.compare == True
    }

    # End of function 'get_dataclass_defaults' #####################################################


# ##################################################################################################
# Classes
# ##################################################################################################

# SamplingParams ###################################################################################

@dataclass(frozen = True)
class SamplingParams:

    """
    Generic class used to specify with which sampling parameters a prompt should be processed.
    """

    # Stopping #####################################################################################

    max_tokens         : int            = field(default = None)
    """
    The number of tokens that may be generated per generation.
    """

    stop               : list[str]      = field(default = None)
    """
    List of character sequences, which if generated stop further generation.
    """

    stop_tokens        : list[int]      = field(default = None)
    """
    List of encoded tokens, which if generated stop further generation.
    """

    include_stop       : bool           = field(default = True)
    """
    Flag that controls whether stop sequences are included in the output.
    """

    ignore_eos         : bool           = field(default = False)
    """
    Flag that controls whether or not generation should continue after the End of Sequence (EOS) 
    token has been generated.
    """


    # Randomization ################################################################################

    temperature        : float          = field(default = 1.0)
    """
    A number between 0.0 and 2.0, with higher values increasing randomization, whereas lower values 
    encourage deterministic output.
    """

    top_p              : float          = field(default = 1.0)
    """
    A number between `0.0` and `1.0`, specifying what is considered a top  percentile, constraining 
    the selection of tokens to tokens with probabilities considered among these top percentiles. For 
    example, when `0.2` is specified, a selection is made from the tokens with probabilities among 
    the top 20 percentiles.
    """

    min_p              : float          = field(default = 0.0)
    """
    A number between `0.0` and `1.0`, specifying what a token's minimal likelihood must be to be 
    considered for selection.
    """

    top_k              : int            = field(default = -1)
    """
    The number of tokens to consider when deciding on the next token when generating.
    """

    logit_bias         : dict[int, int] = field(default = None)
    """
    A dictionary indexed by encoded tokens, each assigned a number between -100 and 100 controlling 
    the likelihood that a specific token is selected or ignored, with a positive number increasing 
    this likelihood, whereas a negative number decreases it.
    """

    seed               : int            = field(default = None)
    """
    A number used to initialize any pseudorandom number generations that the model may used. It can 
    be used to enforce a degree of determinism even when using non-nil temperatures.
    """

    logprobs           : int            = field(default = None)
    """
    The number of tokens to generate per position in the generated sequences, ordered by 
    probability.
    """


    # Repetition ###################################################################################
    
    presence_penalty   : float          = field(default = 0.0)
    """
    A number between -2.0 and 2.0 that controls the likelihood that tokens appear that have not yet 
    appeared in the generated text, whereby a positive  coeffient increases these odds, whereas a 
    negative coefficient decreases them.
    """

    frequency_penalty  : float          = field(default = 0.0)
    """
    A number between -2.0 and 2.0 that controls the number of times a tokens appears in the 
    generated text, whereby a positive coefficient puts a constraint on repetition, whereas a 
    negative coefficient encourages repetition.
    """

    repetition_penalty : float          = field(default = 0.0)
    """
    A number between -2.0 and 2.0 that controls the degree of repetition, taking into account both 
    the generated text and the initial prompt. A positive number encourages usage of new tokens, 
    whereas a negative number favours repetition.
    """

    # End of class 'Params' ########################################################################


# Middleware #######################################################################################

class Middleware:

    """
    Generalised interface for accessing LLMs.
    """

    # ##############################################################################################
    # Methods
    # ##############################################################################################
    
    # exec #########################################################################################
        
    def exec(
        self, 
        query      : str                        = None,
        role       : str                        = None,
        guidance   : str                        = None,
        history    : Thread                     = None,
        context    : dict                       = None,
        params     : SamplingParams             = None
    ) -> tuple[Thread, dict[str, Any]]:
        
        """
        Generate a new message, following up on the message passed using the given guidance and
        sampling parameters.

        Parameters
        ----------

        query
            The query to prompt the LLM with, optionally formatted using Langworks' static DSL.

        role
            The role of the agent stating this query, usually 'user', 'system' or 'assistant'.

        guidance
            Template for the message to be generated, formatted using Langworks' dyanmic DSL.

        history
            Conversational history (thread) to prepend to the prompt.

        context
            Context to reference when filling in the templated parts of the `query`, `guidance` and
            `history`. In case the `Langwork` or the input also define a context, the available 
            contexts are merged. When duplicate attributes are observed, the value is copied from 
            the most specific context, i.e. input context over `Query` context, and `Query` context
            over `Langwork` context.

        params
            Sampling parameters, wrapped by a `SamplingParams` object, specifying how the LLM should 
            select subsequent tokens.
        """

        raise NotImplementedError(
            f"Middleware '{self.__qualname__}' does not implement the 'exec' method;"
        )
    
        # End of method 'exec' #####################################################################

# End of File ######################################################################################


# ##################################################################################################
#
# Title:
#
#   langworks.cache.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, implementing various caches to cache prompt outputs.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
import collections

from dataclasses import (
    dataclass,
    field
)

from typing import (
    Any,
    Callable,
    Hashable
)


# Third party ######################################################################################

# cachetools (Thomas Kemmer)
import cachetools


# ##################################################################################################
# Classes
# ##################################################################################################

# ScoredItem #######################################################################################

@dataclass
class ScoredItem:

    """
    Wrapping class to store any cachable item alongside a score. To be used in conjunction with
    `ScoredCache` or any of its derivates.
    """

    item  : Any   = field(default = None)
    """Item stored."""

    score : float = field(default = 0)
    """Score attributed to the item."""

    # End of class 'ScoredItem' ####################################################################
    

# ScoredCache ######################################################################################

class ScoredCache(cachetools.Cache):

    """
    A regular cache with a scoring mechanism added to it to control overwrites, only overwriting a
    cached value if the new value is associated with a higher score as defined by the user.

    .. note::
        As ScoredCache is implemented on top of Thomas Kemmer's 
        `cachetools <https://github.com/tkem/cachetools>`_, ScoredCache may also be used under
        `the MIT License <https://opensource.org/license/mit>`_ under the copyright of Rosaia B.V.
        (2025).
    """

    # ##############################################################################################
    # Class fundamentals
    # ##############################################################################################

    # __init__ #####################################################################################

    def __init__(
        self,
        maxsize   : int,
        getsizeof : Callable[[Any], int] = None
    ):
        
        """
        Initializes the cache.

        Parameters
        ----------

        maxsize
            Maximum size up until which the cache may grow, after which the cache will start pruning
            items to make room for new keys.
        getsizeof
            Callable used to determine the size of cached items.
        """

        # Initialize inner constructs ##############################################################

        # Initialize super class.
        cachetools.Cache.__init__(self, maxsize, getsizeof)


        # End of '__init__' ########################################################################


    # __getitem__ ##################################################################################

    def __getitem__(
        self, 
        key               : Hashable, 
        __cache_getitem__ : Callable = cachetools.Cache.__getitem__ # Micro optimization
    ) -> Any:
        
        """
        Retrieves item from the cache.

        Parameters
        ----------

        key
            Key of the item to retrieve.
        """

        # Retrieve item from cache, and return wrapped value.
        return __cache_getitem__(self, key).item
    
        # End of '__getitem__' #####################################################################


    # __setitem__ ##################################################################################
    
    def __setitem__(
        self, 
        key               : Hashable, 
        item              : ScoredItem, 
        __cache_getitem__ : Callable    = cachetools.Cache.__getitem__, # Micro optimization
        __cache_setitem__ : Callable    = cachetools.Cache.__setitem__  # Micro optimization
    ):
        
        """
        Adds an item to the cache. 
        
        Unlike regular caches, these items need to be wrapped using a `ScoredItem` object, holding
        the actual item, as well as the score attributed to this item.

        Parameters
        ----------

        key
            Key used to index the item.
        item
            Item to store, wrapped in a `CachedItem`, consisting of an item and a score.
        """

        # Check if an item was stored with the given key.
        if key in self:
            
            # If it was, retrieve the item.
            cached_item = __cache_getitem__(self, key)

            # Compare scores, only proceeding if the new item's score is better.
            if item.score <= cached_item.score:
                return
            
        # Cache the new item.
        __cache_setitem__(self, key, item)

        # End of '__setitem__' #####################################################################

    # End of class 'ScoredCache' ###################################################################


# ScoredLazyLFUCache ###############################################################################

class ScoredLazyLFUCache(ScoredCache):

    """
    A scored Least Frequently Used (LFU) cache, implemented in a 'lazy' manner, only returning cache
    hits for a key if said key has been requested more than a set number of times during
    availability for retrieval. In doing so, this cache effectively samples items before storing
    them.

    .. note::
        As ScoredLazyLFUCache is implemented on top of Thomas Kemmer's 
        `cachetools <https://github.com/tkem/cachetools>`_, ScoredLazyLFUCache may also be used 
        under `the MIT License <https://opensource.org/license/mit>`_ under the copyright of 
        Rosaia B.V. (2025).
    """

    # ##############################################################################################
    # Class fundamentals
    # ##############################################################################################

    # __init__ #####################################################################################

    def __init__(
        self, 
        maxsize   : int, 
        getsizeof : Callable[[Any], int] = None,
        threshold : int = 3
    ):
        
        """
        Initializes the cache.

        Parameters
        ----------

        maxsize
            Maximum size up until which the cache may grow, after which the cache will start pruning
            items to make room for new keys.
        getsizeof
            Callable used to determine the size of cached items.
        threshold
            Minimum number of hits required before the cache starts returning for a key.
        """

        # Initialize inner constructs ##############################################################

        # Initialize super class.
        ScoredCache.__init__(self, maxsize, getsizeof)

        # Requests counter.
        self.__requests = collections.Counter()


        # Initialize attributes ####################################################################

        # Argument passthrough
        self.threshold = threshold


        # End of '__init__' ########################################################################


    # __getitem__ ##################################################################################

    def __getitem__(
        self, 
        key                     : Hashable, 
        __scoredcache_getitem__ : Callable = ScoredCache.__getitem__ # Micro optimization
    ):
        
        """
        Retrieves item from the cache.

        Parameters
        ----------

        key
            Key of the item to retrieve.
        """

        # Attempt to retrieve item from cache.
        item : ScoredItem = __scoredcache_getitem__(self, key)

        # Double check for availability, in case missings are handled differently by any subclasses.
        if key in self:

            # Push up usage counter.
            self.__requests[key] -= 1

            # Check if the threshold was reached.
            if self.__requests[key] > (-self.threshold):

                # If not, pretend the item not to have been cached yet.
                return self.__missing__(key)
            
        # Return the item. 
        return item
    
        # End of '__getitem__' #####################################################################


    # __setitem__ ##################################################################################

    def __setitem__(
        self,
        key                     : Hashable,
        value :                 ScoredItem,
        __scoredcache_setitem__ : Callable = ScoredCache.__setitem__ # Micro optimization
    ):
        
        """
        Adds an item to the cache. 
        
        Unlike regular caches, these items need to be wrapped using a `ScoredItem` object, holding
        the actual item, as well as the score attributed to this item.

        Parameters
        ----------

        key
            Key used to index the item.
        item
            Item to store, wrapped in a `CachedItem`, consisting of an item and a score.
        """
        
        # Store item.
        __scoredcache_setitem__(self, key, value)

        # Initialize counter for object.
        self.__requests[key] += 0
        
        # End of '__setitem__' #####################################################################


    # __delitem__ ##################################################################################

    def __delitem__(
        self,
        key           : Hashable,
        __scoredcache_delitem__ : Callable = ScoredCache.__delitem__ # Micro optimization
    ):
        
        """
        Deletes item from the cache.

        Parameters
        ----------

        key
            Key to delete from the cache.
        """
        
        # Delete from cache.
        __scoredcache_delitem__(self, key)

        # Remove counter for item.
        del self.__requests[key]

        # End of '__delitem__' #####################################################################

    
    # ##############################################################################################
    # Methods
    # ##############################################################################################

    # popitem ######################################################################################

    def popitem(self) -> tuple[Hashable, Any]:

        """
        Remove and return the key of the least frequently used item, as well as the item itself.
        """

        # Attempt to retrieve the key of the least frequently used item.
        try:
            ((key, _),) = self.__requests.most_common(1)
        except:
            raise KeyError("%s is empty" % type(self).__name__) from None
        else:

            # Set requests counter for key beyond the threshold as to ensure that the cache does
            # not report a missing value, when we retrieve the value.
            self.__requests[key] -= self.threshold

            # Return a key-value pair.
            return (key, self.pop(key))
        
        
        # End of method 'popitem' ##################################################################
        
    # End of class 'ScoredLazyLFUCache' ############################################################


# ScoredLazyLRUCache ###############################################################################

class ScoredLazyLRUCache(ScoredCache):

    """
    A scored Least Recently Used (LRU) cache, implemented in a 'lazy' manner, only returning cache
    hits for a key if said key has been requested more than a set number of times during
    availability for retrieval. In doing so, this cache effectively samples items before storing
    them.

    .. note::
        As ScoredLazyLRUCache is implemented on top of Thomas Kemmer's 
        `cachetools <https://github.com/tkem/cachetools>`_, ScoredLazyLRUCache may also be used 
        under `the MIT License <https://opensource.org/license/mit>`_ under the copyright of 
        Rosaia B.V. (2025).
    """

    # ##############################################################################################
    # Class fundamentals
    # ##############################################################################################

    # __init__ #####################################################################################

    def __init__(
        self, 
        maxsize   : int, 
        getsizeof : Callable[[Any], int] = None,
        threshold : int = 3
    ):
        
        """
        Initializes the cache.

        Parameters
        ----------

        maxsize
            Maximum size up until which the cache may grow, after which the cache will start pruning
            items to make room for new keys.
        getsizeof
            Callable used to determine the size of cached items.
        threshold
            Minimum number of hits required before the cache starts returning for a key.
        """

        # Initialize inner constructs ##############################################################

        # Initialize super class.
        ScoredCache.__init__(self, maxsize, getsizeof)

        # Requests counter, which also doubles as a position tracker.
        self.__requests = collections.OrderedDict()


        # Initialize attributes ####################################################################

        # Argument passthrough
        self.threshold : int = threshold


        # End of '__init__' ########################################################################


    # __getitem__ ##################################################################################

    def __getitem__(
        self, 
        key                     : Hashable, 
        __scoredcache_getitem__ : Callable = ScoredCache.__getitem__ # Micro optimization
    ):
        
        """
        Retrieves item from the cache.

        Parameters
        ----------

        key
            Key of the item to retrieve.
        """

        # Attempt to retrieve item from cache.
        item : ScoredItem = __scoredcache_getitem__(self, key)

        # Double check for availability, in case missings are handled differently by any subclasses.
        if key in self:

            # Push up requests counter.
            self.__requests[key] -= 1

            # Update position.
            self.__requests.move_to_end(key)

            # Check if the threshold was reached.
            if self.__requests[key] > (-self.threshold):

                # If not, pretend the item not to have been cached yet.
                return self.__missing__(key)
            
        # Return the item. 
        return item
    
        # End of '__getitem__' #####################################################################


    # __setitem__ ##################################################################################

    def __setitem__(
        self,
        key                     : Hashable,
        value :                 ScoredItem,
        __scoredcache_setitem__ : Callable = ScoredCache.__setitem__ # Micro optimization
    ):
        
        """
        Adds an item to the cache. 
        
        Unlike regular caches, these items need to be wrapped using a `ScoredItem` object, holding
        the actual item, as well as the score attributed to this item.

        Parameters
        ----------

        key
            Key used to index the item.
        item
            Item to store, wrapped in a `CachedItem`, consisting of an item and a score.
        """
        
        # Store item.
        __scoredcache_setitem__(self, key, value)

        # Attempt to move the key to the end of the position tracker.
        try:
            self.__requests.move_to_end(key)
        # If that cannot be done, the tracker has not yet been initialized, and needs to be.
        except:
            self.__requests[key] = 0
        
        # End of '__setitem__' #####################################################################


    # __delitem__ ##################################################################################

    def __delitem__(
        self,
        key           : Hashable,
        __scoredcache_delitem__ : Callable = ScoredCache.__delitem__ # Micro optimization
    ):
        
        """
        Deletes item from the cache.

        Parameters
        ----------

        key
            Key to delete from the cache.
        """
        
        # Delete from cache.
        __scoredcache_delitem__(self, key)

        # Remove counter (and position tracker) for item.
        del self.__requests[key]

        # End of '__delitem__' #####################################################################

    # ##############################################################################################
    # Methods
    # ##############################################################################################

    # popitem ######################################################################################

    def popitem(self) -> tuple[Hashable, Any]:

        """
        Remove and return the key of the least frequently used item, as well as the item itself.
        """

        # Attempt to retrieve the key of least recently used item
        try:
            key = next(iter(self.__requests))
        except:
            raise KeyError("%s is empty" % type(self).__name__) from None
        else:

            # Set requests counter for key beyond the threshold as to ensure that the cache does
            # not report a missing value, when we retrieve the value.
            self.__requests[key] -= self.threshold

            # Return a key-value pair.
            return (key, self.pop(key))
        
        
        # End of method 'popitem' ##################################################################
        
# End of File ######################################################################################


# ##################################################################################################
#
# Title:
#
#   langworks.dsl.environment.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, providing a single access point for rendering templated text.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
from collections import (
    deque
)

from dataclasses import (
    dataclass,
    field
)

from enum import (
    Enum
)

from typing import (
    Any,
    Callable,
    Iterator
)

import warnings

# Runtime
import threading

# Text utilities
import json
import re


# Third party ######################################################################################

# jinja2 (Pallets)
import jinja2


# Local ############################################################################################

# dsl.constraints
from . import constraints

# middleware
from ..middleware import (
    generic,
    vllm
)


# ##################################################################################################
# Globals
# ##################################################################################################

__d_environment__      : jinja2.Environment = None
"""Configured instance of a Jinja2 environment with all DSL extensions enabled."""

__d_environment_lock__ : threading.Lock     = threading.Lock()
"""Lock that governs access to `__d_environment__` global."""

__s_environment__      : jinja2.Environment = None
"""Configured instance of a Jinja2 environment limited to the static subset of the DSL."""

__s_environment_lock__ : threading.Lock     = threading.Lock()
"""Lock that governs access to `__s_environment__` global."""



# ##################################################################################################
# Classes
# ##################################################################################################

# RenderingMode ####################################################################################

class RenderingMode(Enum):

    """
    Specifies how :py:func:`langworks.dsl.environment.render` should process templates, either
    using the 'static' DSL, excluding constructs for guided generation, or using the 'dynamic' DSL,
    including such constructs, among which constraints.
    """

    # Options ######################################################################################

    STATIC  : int = 0
    """
    Renders template in static mode.
    """

    DYNAMIC : int = 1
    """
    Renders template in dynamic mode.
    """

    # End of class 'RenderingMode' #################################################################


# ##################################################################################################
# Functions
# ##################################################################################################

# render ###########################################################################################

def render(
    template : str, 
    context  : dict[str, Any] , 
    mode     : RenderingMode  = RenderingMode.STATIC,
    nested   : bool           = True
) -> Iterator[str | constraints.Constraint]:
    
    """
    Processes the given template set against the context provided, using a Jinja2 environment 
    configured to apply Langwork's DSL.

    Parameters
    ----------

    template
        Template to process.

    context
        Context to reference when processing the template.

    mode
        Whether to apply Langwork's complete 'dynamic' DSL, or the restricted 'static' DSL.

        Also see: :py:class:`langworks.dsl.environment.RenderingMode`.

    nested 
        Flag that determines whether nested Jinja2 constructs are also rendered. If set to `True`
        nested constructs are rendered too, otherwise they are left unprocessed.

    """
    
    # ##############################################################################################
    # Preparation
    # ##############################################################################################

    # Predeclarations ##############################################################################
        
    # string: Holds chunks of text processed by Jinja.
    string : str = None

    # re_jinja_tag: Compiled regular expression to quickly find Jinja opening tags.
    re_jinja_tag : re.Pattern = re.compile(r"\{[\{\%}]")


    # Pre-processing ###############################################################################

    # Split off any text leading up to a Jinja open tag.
    jinja_tag_pos_match : re.Match | None = re_jinja_tag.search(template)

    if jinja_tag_pos_match is None:

        # When no opening tag can be found, the template needs no further processing, and can be
        # returned as-is.
        yield template
        return
    
    else:

        jinja_tag_pos : int = jinja_tag_pos_match.start()

        yield template[:jinja_tag_pos]
        template = template[jinja_tag_pos:]


    # Jinja ########################################################################################

    ### Acquire reference to Jinja environment #####################################################

    # Set-up variable to hold the Jinja environment.
    jinja_env : jinja2.Environment = None

    # Handle loading of dynamic environment:
    if mode == RenderingMode.DYNAMIC:

        # Access relevant globals to access unrestricted environment for rendering.
        global __d_environment__
        global __d_environment_lock__

        # Acquire lock to ensure threadsafe singleton behaviour.
        with __d_environment_lock__:

            # Check if the environment was instantiated before, if not instantiate it.
            if not __d_environment__:

                # Initialize environment.
                __d_environment__ = (
                    jinja2.Environment(

                        cache_size = 128,

                        extensions = [
                            constraints.ChoiceTag,
                            constraints.GenTag,
                            constraints.GrammarTag,
                            constraints.JSONTag,
                            constraints.RegexTag
                        ]
                    )
                )

                # Add globals.
                __d_environment__.globals.update(**{
                    "Params" : generic.SamplingParams,
                    "vllm"   : { "Params": vllm.SamplingParams }
                })

            # Get local reference.
            jinja_env : jinja2.Environment = __d_environment__

    else:

        # Check for unsupported arguments, warning as needed.
        if mode != RenderingMode.STATIC:

            warnings.warn(
                f"Parameter langworks.dsl.environment.render:mode' was set to '{mode}' where either"
                f" 'STATIC' or 'DYNAMIC' was expected; defaulting to 'static';"
            )

        # Access relevant globals to access restricted environment for rendering.
        global __s_environment__
        global __s_environment_lock__

        # Acquire lock to ensure threadsafe singleton behaviour.
        with __s_environment_lock__:

            # Check if the environment was instantiated before, if not instantiate it.
            if not __s_environment__:

                # Initialize environment.
                __s_environment__ = (
                    jinja2.Environment(
                        cache_size = 128,
                    )
                )

            # Get local reference.
            jinja_env : jinja2.Environment = __s_environment__


    ### Create shared context ######################################################################

    # Instantiate a Jinja template based on the given string template.
    jinja_template : jinja2.Template = jinja_env.from_string(template)

    # Create a separate context to hold output variables.
    output_context : dict = {}

    # Using the Jinja template, create a Jinja context object, to be reused by all sub-templates.
    jinja_context = jinja_template.new_context({**context, "output": output_context})


    # ##############################################################################################
    # Processing
    # ##############################################################################################

    # Create a deque to keep track of (sub-)template generators, adding to it the generator of the
    # Jinja template already instantiated.
    queue : deque[Iterator[str]] = deque([jinja_template.root_render_func(jinja_context)])

    # Keep running as long as there are generators that are not exhausted.
    while True:

        # Acquire next string chunk ################################################################

        # Attempt to acquire a generator, breaking if no more generators can be acquired.
        try:
            generator = queue.pop()
        except:
            break

        # Retrieve next string chunk, skipping to the next generator if not more chunks can be
        # retrieved from the current generator.
        string = next(generator, None)

        if string is None:
            continue

        # Put back the generator on the queue, for a next pass. In doing so, the queue is set-up to
        # give precedence to any sub-template generators.
        queue.append(generator)


        # Process chunk ############################################################################

        # Handle constraints.
        if string[:2] == "|%" and mode == RenderingMode.DYNAMIC:

            # Extract identifier of the constraint.
            constraint_hash = string[3:-3]

            # If so, retrieve the constraint object, and yield it.
            constraint : constraints.ConstraintRepr = (
                constraints.cache.get(int(constraint_hash), None)
            )

            # Yield the results.
            if constraint is not None:

                # Decode specs if needed.
                constraint : constraints.Constraint = type(constraint).instantiate(constraint)

                # Add finalizer.
                if constraint.var is not None:
                    constraint.finalizer = lambda x: output_context.update(**{constraint.var: x})

                # Yield result.
                yield constraint

        # Handle other kinds of strings.
        else:

            # Handle nesting as required.
            if nested is True:

                # Split off any text preceeding a Jinja opening tag, preserving computation.
                jinja_tag_pos_match = re_jinja_tag.search(string)

                if jinja_tag_pos_match is not None:

                    # Split up text.
                    jinja_tag_pos = jinja_tag_pos_match.start()

                    yield string[:jinja_tag_pos]
                    string = string[jinja_tag_pos:]

                    # Create a sub-template and accompanying generator, and add it to the queue.
                    jinja_template = jinja_env.from_string(string)
                    queue.append(jinja_template.root_render_func(jinja_context))

                    # Skip to the next cycle.
                    continue

            # Otherwise, return the string unaltered.
            yield string

    # End of function 'render' #####################################################################

# End of File ######################################################################################


# ##################################################################################################
#
# Title:
#
#   langworks.middleware.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, being the initializing script for the middleware sub-module.
#
# ##################################################################################################

# Stub

# End of File ######################################################################################


# ##################################################################################################
#
# Title:
#
#   langworks.langwork.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, implementing the Langwork class, a generic pipework based on
#   the pypework framework specifically designed for working with generative AIs.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
from typing import (
    Sequence,
    TypeVar
)

# System
import logging


# Third party ######################################################################################

# pypework (Rosaia)
from pypeworks import (
    Node,
    Connection,
    Pipework
)


# Local ############################################################################################

from .messages import (
    Thread
)

from .middleware.generic import (
    Middleware
)

from .query import (
    Query
)


# ##################################################################################################
# Classes
# ##################################################################################################

T = TypeVar("T")
R = TypeVar("R")

class Langwork(Pipework[T, R]):

    """
    A `Langwork` is a specialised pypeworks' Pipework allowing for easy access to LLMs.
    """

    # ##############################################################################################
    # Class fundamentals
    # ##############################################################################################

    # __init__ #####################################################################################

    def __init__(
        self,

        system_msg     : Thread | None                   = None,
        ignore_history : bool                            = False,
        context        : dict | None                     = None,
        ignore_context : bool | Sequence[str] | None     = None,
        middleware     : Middleware | None               = None,

        processes      : int | None                      = None,
        logger        : logging.Logger          = None,
        ignore_errors : Sequence[BaseException] = None,

        connections    : list[Connection]                = [],
        **nodes        : Query | type["Langwork"] | Node
    ):
        
        """
        Sets up a `Langwork` object, a generic pipework based on the pypeworks framework designed
        specifically to query generative AIs.

        Parameters
        ----------

        system_msg
            System messages to prepend to any history provided by a `Query`. When the input 
            specifies a specific history, the system messages are *not* prepended.

        ignore_history
            Flag that signals whether any history provided by `Query` or input should be ignored.
            By default these histories are taken into account. If set to `True`, these histories are
            ignored, and each prompt starts with the given system messages, as if starting a new 
            thread.

        context
            Context used by Langworks' DSL to fill in any templated parts of any queries, history 
            or guidance included in the prompt passed to the LLM. In case the `Query` or `Langwork` 
            also define a context, the available contexts are merged. When duplicate attributes are
            observed, the value is copied from the most specific context, i.e. input context over 
            `Query` context, and `Query` context  over `Langwork` context.

        ignore_context
            Controls how contexts passed by argument are handled. By default contexts specified by
            `Langwork` may be overwritten by :py:class:`langworks.query.Query` or contexts passed
            by argument. However, when `ignore_context` is set, `Langwork` blocks this behaviour,
            instead maintaining the context as specified by `Langwork`. When `ignore_context` is set
            to `True` none of the keys specified by `Langwork` may be overwritten. Alternatively, if
            a sequence of keys is passed, only the keys specified are protected from overwriting. 

        middleware
            The middleware used to connect to the LLM. If not defined, the framework will try to use
            any middlewares included in the specifications of individual `Query` objects.

        processes
            The number of worker processes to use to operate the `Langwork`. By default this number 
            is equal to the number of logical CPUs in the system.

        connections
            Specification of connections in the langwork. See pypework's documentation for further 
            information.

        logger
            Logger compatible with Python's 
            `logging <https://docs.python.org/3/library/logging.html>`_ module, providing control 
            over the registration of events occuring within the langwork.

        ignore_errors
            Sequence of errors (and exceptions), which if raised, are ignored by the Pipework,
            allowing it to continue execution.

        nodes
            Nodes to embed in the langwork, including 'Query' and 'Langwork' objects.
        """

        # Initialize attributes ####################################################################

        # Argument passthrough
        self.system_msg     : Thread | None        = system_msg
        self.ignore_history : bool                 = ignore_history
        self.context        : dict | None          = context
        self.ignore_context : bool | Sequence[str] = ignore_context 
        self.middleware     : Middleware | None    = middleware

        # Finish initialization of any queries passed.
        for node in nodes.values():
            
            if isinstance(node, Langwork):

                node.system_msg = node.system_msg or system_msg
                node.context    = {**(node.context or dict()), **(context or dict())}
                node.middleware = node.middleware or middleware
        

        # ##########################################################################################
        # Delegation to super class
        # ##########################################################################################

        # Initialize inner pipework.
        self.pipework = (
            super().__init__(
                processes     = processes, 
                logger        = logger,
                ignore_errors = ignore_errors,
                connections   = connections, 
                **nodes
            )
        )

        # End of method '__init__' #################################################################

    # End of class 'Langwork' ######################################################################
        
# End of File ######################################################################################


# ##################################################################################################
#
# Title:
#
#   langworks.middleware.vllm.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, implementing middleware to leverage vLLM.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
from dataclasses import (
    dataclass,
    field
)

import typing
from typing import (
    Any,
    Callable,
    Literal
)

import warnings

# System
from threading import (
    Lock
)

# Text utilities
import json

# Utilities
import math
import statistics


# Third-party ######################################################################################

# OpenAI API (OpenAI)
try:
    import openai
except:
    pass

# transformers
try:
    from transformers import (
        AutoTokenizer
    )
except:
    pass


# Local ############################################################################################

from .. import (
    auth,
    caching,
    dsl
)

from ..messages import (
    Thread
)

from . import (
    generic
)


# ##################################################################################################
# Classes
# ##################################################################################################

# Params ###########################################################################################

@dataclass(frozen = True)
class SamplingParams(generic.SamplingParams):

    """
    Specifies additional sampling parameters to be used when passing a prompt to vLLM.
    """

    # Beam search ##################################################################################

    use_beam_search               : bool                      = field(default = False)
    """
    Flag specifying whether or not to apply beam search.
    """

    best_of                       : int                       = field(default = None)
    """
    The number of options to explore when applying beam search.
    """

    length_penalty                : float                     = field(default = 1.0)
    """
    A coeffient applied during beam search, controlling the length of the generated text.
    """

    early_stopping                : bool | Literal["never"]   = field(default = False)
    """
    Controls stopping behaviour during beam search. This parameter takes three values:

    - `True`, whereby generation is stopped when `best_of` candidates are acquired;
    - `False`, whereby generation continues as long as the probability remains high that better
      candidates may be found;
    - `"never"`, whereby generation continues as long as better candidates may be found;
    """

# Monkey patch in '__defaults__', a quick lookup map for default values ############################

SamplingParams.__defaults__ = generic.get_dataclass_defaults(SamplingParams)

# End of class 'Params' ############################################################################


# vLLM #############################################################################################

class vLLM(generic.Middleware):

    """
    `vLLM <https://docs.vllm.ai/en/stable/>`_ is an open source library for efficiently serving
    various LLMs using self-managed hardware. Langworks' vLLM middleware provides a wrapper to 
    access these LLMs using vLLM's server API.
    """

    # ##############################################################################################
    # Attributes
    # ##############################################################################################

    # Private ######################################################################################

    # Cache of chat templates indexed by model name.
    __templates__ : dict[str, str] = dict()

    # Cache of special tokens indexed by model name.
    __special_tokens__ : dict[str, str] = dict()


    # ##############################################################################################
    # Class fundamentals
    # ##############################################################################################

    # __init__ #####################################################################################

    def __init__(
            
            self, 

            # Connection
            url           : str,
            model         : str,
            authenticator : auth.Authenticator | None = None,
            timeout       : int                       = 5,
            retries       : int                       = 2,
            
            # Configuration
            params        : SamplingParams            = None,
            output_cache  : caching.ScoredCache       = None
            
    ):
        
        """
        Initializes the vLLM middleware.

        Parameters
        ----------

        Connection
        ^^^^^^^^^^

        url
            URL of the vLLM-instance where the model may be accessed.

        model
            Name of the model as used in the Hugging Face repository.

        authenticator
            :py:class:`Authenticator` that may be used to acquire an authentication key when the 
            vLLM-instance requests for authentication.

        timeout
            The number of seconds the client awaits the response of the vLLM-instance.

        retries
            The number of times the client tries to submit the same request again *after* a timeout.


        Configuration
        ^^^^^^^^^^^^^

        params
            Default sampling parameters to use when processing a prompt, specified using an instance
            of `SamplingParams`.

        output_cache
            Cache to use for caching previous prompt outputs. Should be a `ScoredCache` or subclass.

        """

        # Initialize attributes ####################################################################

        # Argument passthrough
        self.url               : str                       = url
        self.model             : str                       = model
        self.authenticator     : auth.Authenticator | None = authenticator
        self.timeout           : int                       = timeout
        self.retries           : int                       = retries
        self.params            : SamplingParams            = params
        self.output_cache      : caching.ScoredCache       = output_cache

        # Pre-declare attributes
        self.client            : openai.OpenAI             = None

        self.chat_template     : str                       = None
        self.special_tokens    : dict[str, str]            = None

        self.output_cache_lock : Lock                      = Lock()

        # End of '__init__' ########################################################################


    # ##############################################################################################
    # Methods
    # ##############################################################################################
        
    # exec #########################################################################################
        
    def exec(
        self, 
        query      : str                        = None,
        role       : str                        = None,
        guidance   : str                        = None,
        history    : Thread                     = None,
        context    : dict                       = None,
        params     : SamplingParams             = None
    ) -> tuple[Thread, dict[str, Any]]:
        
        """
        Generate a new message, following up on the message passed using the given guidance and
        sampling parameters.

        Parameters
        ----------

        query
            The query to prompt the LLM with, optionally formatted using Langworks' static DSL.

        role
            The role of the agent stating this query, usually 'user', 'system' or 'assistant'.

        guidance
            Template for the message to be generated, formatted using Langworks' dyanmic DSL.

        history
            Conversational history (thread) to prepend to the prompt.

        context
            Context to reference when filling in the templated parts of the `query`, `guidance` and
            `history`. In case the `Langwork` or the input also define a context, the available 
            contexts are merged. When duplicate attributes are observed, the value is copied from 
            the most specific context, i.e. input context over `Query` context, and `Query` context
            over `Langwork` context.

        params
            Sampling parameters, wrapped by a `SamplingParams` object, specifying how the LLM should 
            select subsequent tokens.
        """
        
        # ##########################################################################################
        # Prepare initial prompt
        # ##########################################################################################

        # Render initial messages.
        messages : Thread = [
            # Construct new message.
            {
                # Maintaining original role.
                "role"   : message.get("role", None),
                # Replacing the content with the rendered variant.
                "content": (
                    # Join rendered chunks.
                    "".join(
                        # Pass to static DSL environment.
                        dsl.render(
                            message.get("content", None), context, mode = dsl.RenderingMode.STATIC
                        )
                    )
                )
            }
            # Process each message separately.
            for message in (
                # Append query to the history.
                (history or []) + [{"content": query or "", "role": role or "user"}]
            )
        ]

        # Convert thread to prompt ready for further generation.
        prompt : str = (
            # Join rendered chunks.
            "".join(
                # Pass to static DSL environment.
                dsl.render(

                    # Retrieve chat template.
                    self.get_chat_template(), 

                    # Populate the context.
                    dict(
                        **context,
                        messages              = messages,
                        add_generation_prompt = True,
                        **self.get_special_tokens()
                    ),

                    # Only the template itself needs to be rendered.
                    nested = False,

                    # Set mode to static.
                    mode = dsl.RenderingMode.STATIC
                )
            )
        )


        # ##########################################################################################
        # Prepare rendering
        # ##########################################################################################

        # Get connection with vLLM instance ########################################################

        # Check if the connection still needs to be initialized.
        if self.client is None:

            # Retrieve token if an authenticator was passed.
            token : str | None = None

            if self.authenticator is not None:
                
                details : auth.AuthenticationDetails = self.authenticator(self)

                if details is not None:
                    token = details.token

                if token is None:

                    warnings.warn(
                        f"Authenticator passed could not provide authentication details; defaulting"
                        " to token 'EMPTY';"
                    )
        
            # Import OpenAI client, and use it to make a connection to the vLLM instance.
            self.client = openai.OpenAI(base_url = self.url, api_key = token or "EMPTY")


        # Combine default sampling parameters ######################################################

        # Combine sampling parameters specified at different levels.
        params = (

            # Do NOT create a new SamplingParams object, maintaining instead a dictionary as to
            # minimize the required computation when mixin constraint-level sampling parameters
            # later on.
            {

                # Middleware
                **{
                    key: value 
                    for key, value in 
                    (self.params.__dict__ if self.params is not None else {}).items()
                    if value != SamplingParams.__defaults__.get(key, None)
                },

                # Input
                **{
                    key: value 
                    for key, value in 
                    (params.__dict__ if params is not None else {}).items()
                    if value != SamplingParams.__defaults__.get(key, None)
                }
            }

        )


        # ##########################################################################################
        # Rendering
        # ##########################################################################################

        # Define intermediate variables ############################################################

        # Variable to buffer the content of the message being generated.
        msg_content : str = ""

        # Variable to hold generated variables.
        vars : dict[str, Any] = dict()


        # Main logic ###############################################################################

        # Process the guide tag by tag.
        for chunk in dsl.render(
            guidance or "{% gen %}", context, mode = dsl.RenderingMode.DYNAMIC
        ):

            # Differentiate constraints and non-constraints ########################################

            # Add any strings to the buffer and prompt, and skip until a constraint is found.
            if isinstance(chunk, str):
                
                msg_content += chunk
                prompt += chunk

                continue

            # Check if a constraint was passed.
            elif not isinstance(chunk, dsl.Constraint):

                raise ValueError(
                    f"Object of type '{type(chunk)}' was passed where 'str' or subtype of 'dsl."
                    "constraints.Constraint was expected;"
                )


            # Mixin final sampling parameters ######################################################
            
            # Combine sampling parameters stored in the constraint with default constraints.
            prompt_params : SamplingParams = (

                SamplingParams(**{

                    **{
                        key: value for key, value in params.items()
                    },

                    **{
                        key: value 
                        for key, value in 
                        (chunk.params.__dict__ if chunk.params is not None else {}).items()
                        if value != SamplingParams.__defaults__.get(key, None)
                    }
                })

            )


            # Generate response ####################################################################

            # Predefine variable to hold text response.
            txt : str | None = None

            # Check cache to see if a response needs to be generated.
            if self.output_cache is not None:

                # Generate hash for the request.
                request_hash : int = hash(prompt + self.model + str(prompt_params))

                # Access cache, and attempt to retrieve response.
                with self.output_cache_lock:

                    try:
                        txt = self.output_cache[request_hash]
                    except:
                        pass

            # Only retrieve response, if a response was not previously cached.
            if txt is None:

                # Prepare an object, holding all request arguments.
                request : dict = dict(

                    model              = self.model,
                    prompt             = prompt,
                    timeout            = self.timeout,

                    max_tokens         = prompt_params.max_tokens,
                    stop               = prompt_params.stop,

                    temperature        = prompt_params.temperature,
                    top_p              = prompt_params.top_p,      
                    logit_bias         = prompt_params.logit_bias,
                    seed               = prompt_params.seed,
                    logprobs           = prompt_params.logprobs,

                    presence_penalty   = prompt_params.presence_penalty,
                    frequency_penalty  = prompt_params.frequency_penalty,

                    extra_body = {

                        "stop_token_ids"             : prompt_params.stop_tokens,
                        "include_stop_str_in_output" : prompt_params.include_stop,
                        "ignore_eos"                 : prompt_params.ignore_eos,

                        "min_p"                      : prompt_params.min_p,
                        "top_k"                      : prompt_params.top_k,

                        #"repetition_penalty"         : prompt_params.repetition_penalty,

                        "use_beam_search"            : prompt_params.use_beam_search,
                        "best_of"                    : prompt_params.best_of,
                        "length_penalty"             : prompt_params.length_penalty,
                        "early_stopping"             : prompt_params.early_stopping,
                    
                        **({
                            (
                                "guided_regex"        if isinstance(chunk,  dsl.Regex)
                                else "guided_json"    if isinstance(chunk,  dsl.JSON)
                                else "guided_choice"  if isinstance(chunk,  dsl.Choice)
                                else "guided_grammar" if isinstance(chunk,  dsl.Grammar)
                                else ValueError(f"Unknown constraint '{chunk}' was passed;")
                            ): chunk.spec
                        } if not isinstance(chunk, dsl.Gen) else {})
                    }

                )

                # Query the vLLM-instance.
                attempt : int = -1
                response : openai.types.Completion = None

                while attempt < self.retries:

                    try:
                        response = self.client.completions.create(**request)

                    except openai.APITimeoutError:

                        attempt += 1
                        continue

                    break

                if attempt >= self.retries:

                    raise TimeoutError(
                        f"Connection with '{self.url}' timed out despite {self.retries} retries;"
                    )

                # Get response text.
                txt : str = response.choices[0].text

                # Cache if needed.
                if self.output_cache is not None:

                    # Retrieve logprobs, to be used as the score for comparison.
                    logprobs = response.choices[0].logprobs

                    if logprobs is not None:
                        logprobs = logprobs.token_logprobs

                    logprobs = logprobs or []

                    # Convert to average p-value
                    p = math.exp(statistics.fmean(logprobs)) if len(logprobs) > 0 else 0.0

                    # Add to cache.
                    with self.output_cache_lock:
                        self.output_cache[request_hash] = caching.ScoredItem(txt, p)


            # Process response #####################################################################

            # Append response to message.
            msg_content += txt

            # Append response to prompt.
            prompt += txt

            # Store response if needed.
            if isinstance(chunk, dsl.constraints.Constraint) and chunk.var is not None:

                # Handle JSON objects.
                if isinstance(chunk, dsl.constraints.JSON):
                    vars[chunk.var] = json.loads(txt)

                # Handle other objects.
                else:
                    vars[chunk.var] = txt

                # Invoke finalize.
                if chunk.finalizer is not None:
                    chunk.finalizer(txt)


        # ##########################################################################################
        # Return result
        # ##########################################################################################

        return (
            messages + [{"role": "assistant", "content": msg_content}],
            vars
        )
    
        # End of method 'exec' #####################################################################

        
    # get_chat_template ############################################################################
    
    def get_chat_template(self) -> str:

        """
        Retrieves chat template associated with the model to which this middleware provides access.
        """

        # Attempt to retrieve from cache ###########################################################

        chat_template = vLLM.__templates__.get(self.model, None)

        if chat_template is not None:
            return chat_template
        
        # Add to cache and return ################################################################## 

        # Delegate
        self.__get_chat_format_arguments__()

        # Return chat template.
        return vLLM.__templates__.get(self.model, None)
    
        # End of method 'get_chat_template' ########################################################
        

    # get_special_tokens ###########################################################################

    def get_special_tokens(self) -> dict[str, str]:

        """
        Retrieves special tokens associated with the model to which this middleware provides access.
        """

        # Attempt to retrieve from cache ###########################################################

        special_tokens = vLLM.__special_tokens__.get(self.model, None)

        if special_tokens is not None:
            return special_tokens
        
        # Add to cache and return ################################################################## 

        # Delegate
        self.__get_chat_format_arguments__()

        # Return special tokens.
        return vLLM.__special_tokens__.get(self.model, None)
    
        # End of method 'get_chat_template' ########################################################


    # __get_chat_format_arguments__ ################################################################

    def __get_chat_format_arguments__(self):

        """
        Retrieve and cache arguments used to format a human readable thread as a prompt that can be
        processed by the specified model.
        """

        # Retrieve and cache arguments #############################################################

        # Access the tokenizer for the given model.
        tokenizer = AutoTokenizer.from_pretrained(self.model)

        # Retrieve the chat template from the tokenizer, and cache it.
        vLLM.__templates__[self.model]  = (
            tokenizer.chat_template or tokenizer.default_chat_template
        )

        # Retrieve special tokens from the tokenizer, and cache them.
        vLLM.__special_tokens__[self.model] = tokenizer.special_tokens_map or {} 


        # End of method 'get_chat_template' ########################################################

# End of File ######################################################################################


# ##################################################################################################
#
# Title:
#
#   langworks.query.py
#
# License:
#
#   Copyright 2025 Rosaia B.V.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except 
#   in compliance with the License. You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software distributed under the 
#   License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing permissions and 
#   limitations under the License.
#
#   [Apache License, version 2.0]
#
# Description: 
#
#   Part of the Langworks framework, implementing the Query class, wrapping the logic for
#   accessing an external generative AI.
#
# ##################################################################################################

# ##################################################################################################
# Dependencies
# ##################################################################################################

# Python standard library ##########################################################################

# Fundamentals
from typing import (
    Any,
    Sequence
)


# Third-party ######################################################################################

# pypeworks (Rosaia)
from pypeworks import (
    Node
)

from pypeworks.typing import (
    Args,
    Param
)


# Local ############################################################################################

from .messages import (
    Thread
)

from .middleware.generic import (
    Middleware,
    SamplingParams
)


# ##################################################################################################
# Classes
# ##################################################################################################

# Query ############################################################################################

class Query(Node):

    """
    A `Query` is a specialised Pypeworks Node wrapping a templatable prompt that may be passed to a
    LLM.
    """

    # ##############################################################################################
    # Class fundamentals
    # ##############################################################################################

    # __init__ #####################################################################################

    def __init__(
        self,
        query                 : str,
        role                  : str | None                  = None,
        guidance              : str | None                  = None,
        history               : Thread | None               = None,
        ignore_history        : bool                        = False,
        context               : dict | None                 = None,
        ignore_context        : bool | Sequence[str] | None = None,
        include_in_output     : Sequence[str] | None        = None,
        middleware            : Middleware | None           = None,
        **kwargs
    ):
        
        """
        Sets up a `Query` object, wrapping a query stated by a human, as well as any guidance to
        pass to a LLM when prompted with this query. A `Query` object can be used standalone or in
        conjunction with a `Langwork`.

        Parameters
        ----------

        query
            The query to prompt the LLM with, optionally formatted using Langworks' static DSL.

        role
            The role of the agent stating this query, usually 'user', 'system' or 'assistant'.

        guidance
            Template for the message to be generated, formatted using Langworks' dynamic DSL.

        history 
            Conversational history (thread) to prepend to the prompt when no such history is
            available from either the input or the `Langwork`.

        ignore_history
            Flag that signals whether any preceeding conversational history should be ignored or
            not. By default this history is taken into account. If not, effectively a new 
            conversation is started, using the contents of `history` as specified for this node or
            the wrapping langwork.

        context
            Context to reference when filling in the templated parts of the `query`, `guidance` and
            `history`. In case the `Langwork` or the input also define a context, the available 
            contexts are merged. When duplicate attributes are observed, the value is copied from 
            the most specific context, i.e. input context over `Query` context, and `Query` context
            over `Langwork` context.

        ignore_context
            Controls how contexts passed by argument are handled. By default any context passed
            overwrites the keys set by `Query` and :py:class:`Langwork`. However, when 
            `ignore_context` is set, `Query` blocks this overwriting behaviour, instead maintaining
            the context as previously specified. When `ignore_context` is set to `True` none of the 
            priorly specified keys may be overwritten. Alternatively, if a sequence of keys is
            passed, only those keys passed are protected from overwriting.

        include_in_output
            Controls what items from contexts processed are included in the output generated by
            :py:meth:`Query.exec`. By default, when set to `None`, only a limited context is 
            returned, consisting of the context gathered from the input and the result. However,
            optionally a list of keys may be specified of keys to include in the output, including
            those added through `context`. 
        
        middleware
            The middleware used to connect to the LLM. If not defined, the Langwork's middleware is 
            used instead.

        kwargs
            Any additional arguments to pass to the underlying Pypeworks `Node`.
        """

        # Initialize attributes ####################################################################

        # Argument passthrough
        self.query                 : str                         = query
        self.role                  : str | None                  = role
        self.guidance              : str | None                  = guidance
        self.history               : Thread | None               = history
        self.ignore_history        : bool                        = ignore_history
        self.context               : dict | None                 = context
        self.ignore_context        : bool | Sequence[str] | None = ignore_context
        self.include_in_output     : Sequence[str] | None        = include_in_output
        self.middleware            : Middleware | None           = middleware

        # Pre-initialization
        from .langwork import Langwork
        self.parent                : Langwork = None # Assigned by Langwork
    

        # Invoke super class #######################################################################

        super().__init__(self.exec, **kwargs)


        # End of method '__init__' #################################################################
    

    # ##############################################################################################
    # Methods
    # ##############################################################################################
    
    def exec(
        self, 
        input   : Any | None            = None,
        history : Thread | None         = None, 
        context : dict | None           = None, 
        params  : SamplingParams | None = None, 
        **kwargs
    ) -> Args[Param[Thread, "history"], Param[dict, "context"]]:

        """
        Takes the given history and context, combines it with the histories and contexts included as
        part of the specification of the `Node` and `Langwork`, and forwards it to the middleware, 
        which feeds it to a LLM to generate a response. When a response can be required, a new 
        thread - `history` - is drawn up, consisting of processed historic messages, as well as the 
        message generated by the LLM. Alongside, a new `context` is returned, containing any named 
        generated content as specificied by the DSL included in the node's `guidance`.

        Parameters
        ----------

        input
            Any anonymous input passed to this node, to be included in the context under the key
            'input'.

        history
            Conversational history (thread) to prepend to the prompt. This history supersedes any
            history specified by the `Node` or `Langwork`, unless either was instructed to ignore
            the histories provided by inputs or nodes (`ignore_history=True`).

        context
            Context used by Langwork's DSL to fill in any templated parts of any queries, history
            or guidance included in the prompt passed to the LLM. In case the `Node` or `Langwork`
            also define a context, the available contexts are merged. When duplicate attributes are
            observed, the value is copied from the most specific context, i.e. input context over 
            `Query` context, and `Query` context  over `Langwork` context.

        params
            Sampling parameters, wrapped by a `SamplingParams` object, specifying how the LLM should
            select subsequent tokens.

        kwargs 
            Any key-value arguments to be passed as additional context for use in Langworks' DSL.
        
        """

        # ##########################################################################################
        # Preparation
        # ##########################################################################################

        # Pre-define variables #####################################################################

        # complete_context: context used by Langwork's DSL.
        complete_context : dict[str, Any] = None

        # local_history: list of preceeding messages to pass to the model.
        local_history : Thread = None


        # Combine arguments with predefined values #################################################

        # Merge contexts provided by the langwork, the node and the input.
        complete_context = {
            # parent (Langwork), ignorables 
            **(
                self.parent.context if (
                    self.parent is not None and self.parent.context is not None
                ) else {}
            ),
            # self (Query), ignorables
            **(self.context if self.context is not None else {}),
            # context
            **(context or {}),
            # kwargs
            **kwargs,
            # input
            **({"input": input} if input else dict()),
            # self (Query), protected
            **({
                key: self.context[key] 
                for key in (
                    [] if self.ignore_context in [False, None]
                    else self.context.keys() if isinstance(self.ignore_context, bool) # True
                    else self.ignore_context
                ) 
                if key in self.context
            } if self.context is not None else {}),
            # parent (Langwork), protected
            **({
                key: self.parent.context[key] 
                for key in (
                    [] if self.parent.ignore_context in [False, None]
                    else self.parent.context.keys() if isinstance(self.parent.ignore_context, bool)
                    else self.parent.ignore_context
                ) 
                if key in self.context
            } if self.parent is not None and self.parent.context is not None else {})
        }

        # Select most specific history available.
        local_history = (
            # Default to system messages if Langwork was set to ignore histories.
            (
                self.parent.system_msg or []
            )
            if self.parent is not None and self.parent.ignore_history is True
            # Otherwise pass history provided by Query or input.
            else (
                # Return history provided by Query if it was set to ignore input histories, or if
                # the input provided no history.
                (
                    # Prepend system messages.
                    (
                        self.parent.system_msg 
                        if self.parent is not None 
                        else []
                    )
                    # Add messages from history provided by Query.
                    + (
                        self.history or []
                    )
                )
                if self.ignore_history or history is None
                # Otherwise default to history provided by input.
                else (
                    history or []
                )
            )
        )


        # Get handle on middleware #################################################################

        # Get middleware
        middleware : Middleware = (
            self.middleware
            or (self.parent.middleware if self.parent is not None else None)
        )

        # Raise error if no handle was obtained.
        if middleware is None:
            
            raise RuntimeError(
                "Neither the query nor the langwork specified middleware to communicate with a LLM;"
                " input cannot be processed;"
            )
        

        # ##########################################################################################
        # Forward to middleware, and return result.
        # ##########################################################################################

        # Get response.
        result_history, result_context = (
            middleware.exec(
                self.query, self.role, self.guidance, local_history, complete_context, params
            )
        )

        # Return result.
        return (

            # history
            result_history,
            
            # context
            {
                # Select inputs to include in resulting context.
                **{
                    key: complete_context[key]
                    for key in {
                        "input", 
                        *(kwargs.keys()), 
                        *(self.include_in_output if self.include_in_output is not None else ())
                    }
                    if key in complete_context
                },

                # Results always overwrite all prior context inputs.
                **result_context
            }

        )
    

        # End of method 'exec' #####################################################################

# End of File ######################################################################################